//*********************************************************************
// Copyright (C) 2005,  Lucian Constantin <struct_bylighting@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//*********************************************************************

#include "dictionaries.h"
#include "progressbar.h"
#include "qwerty.h"
#include "qtutils.h"
#include "spell.h"
static bool isUtf8Valid(const char *str);

// defined in unmunch.c
extern "C" int load_dictionary(const char *aff, const char *dict);

// language definitions
static LangDef language[MAX_LANG_DEF] = { // should match language defs in global.h
	{"English (US)", "en_US"},
	{"Español", "es_ES"},
	{"Français", "fr_FR"},
	{"Deutsch", "de_DE"},
	{"Româneşte (cedilla)", "ro_RO"},
	{"Româneşte (commabelow)", "ro_RO"}
};

QString getLangName(int id) {
	ASSERT(id >= 0 && id < MAX_LANG_DEF);
	return QString::fromUtf8(language[id].name);
}

QString getLangDict(int id) {
	ASSERT(id >= 0 && id < MAX_LANG_DEF);
	return QString::fromUtf8(language[id].dict);
}

int getLangId(QString name) {
	for (int i = 0; i < MAX_LANG_DEF; i++) {
		if (QString::fromUtf8(language[i].name) == name)
			return i;

	}
	ASSERT(0);
	return LANG_EN_US;
}

static DbDict *hptr = 0;
static int cnt = 0;
static bool print = false;
static char *trylist = 0;
// unmunch callback: add word to dictionary
extern "C" void add_word_callback(char *word) {
	ASSERT(word != 0);
	if (!hptr->findString(word, true))
		hptr->setString(word, true);
	if (++cnt % 20000 == 0) {
		qApp->processEvents();
		if (print && cnt % 100000 == 0)
			fprintf(stderr, "%d\n", cnt);
	}
}
// unmunchc callback: add trylist
extern "C" void add_trylist_callback(char *word) {
	ASSERT(word != 0);
	ASSERT(trylist == 0);
	
	trylist = new char[strlen(word) + 1];
	strcpy(trylist, word);
}

Dictionaries::Dictionaries(Editor *e, bool readmaster, bool localread, bool localsave): QObject(), 
	e_(e), hptr_(0), iptr_(0), uptr_(0), readmaster_(readmaster), localread_(localread), localsave_(localsave), trylist_(0) {
	LOG("Dictionaries() readmaster %s\n", (readmaster)? "true":"false");
	
	if (e == 0)
		print = true; // count words on stderr as they are loaded
	else
		print = false;
		
	hptr_ = new DbDict();
	if (readmaster)
		readMaster();

	iptr_ = new DbUser();
	uptr_ = new DbUser();
	if (localread)
		readUser();
}

Dictionaries::~Dictionaries() {
	LOG("save dictionaries");

	// save new local dictionaries
	if (localsave_) {
		char *homedir = getenv("HOME");
		if (homedir == 0) {
			fprintf(stderr, "Error: HOME environment variable not found\n");
			exit(1);
		}
		unsigned len = strlen(homedir) + strlen("/.rospell/dictionary.user");
		char *fname = new char[len + 1];
		memset(fname, '\0', len + 1);
		strcpy(fname, homedir);
		strcat(fname, "/.rospell/dictionary.user");
		uptr_->putFile(fname);
		delete[] fname;
	}
	delete hptr_;
	delete iptr_;
	delete uptr_;
	if (trylist_ != 0)
		delete [] trylist_;
}

void Dictionaries::readMaster() {
	if (hptr_ == 0)
		hptr_ = new DbDict();
	hptr = hptr_;
	cnt = 0;
	QString dic = QString(INSTALL_PLACE) + getLangDict(Global::instance()->lang) + QString(".dic");
	QString aff = QString(INSTALL_PLACE) + getLangDict(Global::instance()->lang) + QString(".aff");
	int range = 60;

	ProgressBar *progress = 0;
	if (e_ != 0 && e_->qvb_ != 0) {
		progress = new ProgressBar(e_->qvb_, (const char *) tr("Loading spelling dictionary"));
		progress->show();
		progress->start(range);
		qApp->processEvents();
	}



	if (load_dictionary((const char *) aff, (const char *) dic))
		LOG("Starting with an empty master dictionary: %s/%s", (const char *) aff, (const char *) dic);
	else {
		LOG("%s, %s dictionary loaded\n", (const char *) aff, (const char *) dic);
		// grab the trylist if any
		if (trylist != 0) {
			trylist_ = trylist;
			trylist = 0;
		}
	}
	if (progress != 0)
		delete progress;
}

void Dictionaries::readUser() {
	LOG("readHomeDir");
	// find user file
	// save new dictionary
	char *homedir = getenv("HOME");
	if (homedir == 0) {
		fprintf(stderr, "Error: HOME environment variable not found\n");
		exit(1);
	}

	int len = strlen(homedir) + strlen("/.rospell/dictionary.user");
	char *fname = new char[len + 1];
	memset(fname, '\0', len + 1);
	strcpy(fname, homedir);
	strcat(fname, "/.rospell/dictionary.user");
	char *fstr = UTILS::read_file_malloc(fname);
	if (fstr != 0) {
		LOG("readHomeDir::reading dictionary.user");
		Action *aptr;
		aptr = new ActionAddAllUser();
		spell(this, aptr, fstr);
		delete aptr;
	}
	delete[] fname;
	delete[] fstr;

	Stats::instance()->reset();
	LOG("HOME file was read\n");
}

bool Dictionaries::findString(const char *active_word) {
	if (hptr_ == 0 || iptr_ == 0 || uptr_ == 0) {
		ASSERT(0);
		return 0;
	}
	
	if (!hptr_->findString(active_word, true) &&	// if not in dictionary, fixed sintax
	    !hptr_->findString(active_word) &&	// if not in dictionary
	    !uptr_->findString(active_word) &&	// if not in user dictionary
	    !uptr_->findString(active_word, true) &&	// if not in user dictionary, fixed sintax
	    !iptr_->findString(active_word) &&	// if not in ignore dictionary
	    !iptr_->findString(active_word, true))	// if not in ignore dictionary, fixed sintax
		
		return false; // word not found

	return true; // word found
}

bool Dictionaries::suggestFlip(const char *active_word, char *dest1, char *dest2, char *dest3) {
	ASSERT(active_word != 0);
	ASSERT(dest1 != 0);
	ASSERT(dest2 != 0);
	ASSERT(dest3 != 0);
	*dest1 = '\0';
	*dest2 = '\0';
	*dest3 = '\0';

	bool retval = false;
	char newword[100 + 1];
	strncpy(newword, active_word, 100);
	unsigned char *ptr = (unsigned char *) newword;

	// warning: utf8 codes are messed up!!!
	// lucian: provide an implementation to suport variable char length for utf8
	while (*ptr != 0 && *(ptr + 1) != 0) {
		if (!(*ptr & 0x80  || *(ptr + 1) & 0x80)) {

			char tmp = *ptr;
			*ptr = *(ptr + 1);
			*(ptr + 1) = tmp;
			if (isUtf8Valid(newword)) {
				if (hptr_->findString(newword, true) ||
				    hptr_->findString(newword)) {
					retval = true;
					if (*dest1 == '\0')
						strcpy(dest1, newword);
					else if (*dest2 == '\0' && strcmp(dest1, dest2) != 0)
						strcpy(dest2, newword);
					else if (*dest3 == '\0' && strcmp(dest1, dest3) != 0 && strcmp(dest2, dest3) != 0) {
						strcpy(dest3, newword);
						return true;
					}
				}
			}

			tmp = *ptr;
			*ptr = *(ptr + 1);
			*(ptr + 1) = tmp;
		}
		ptr++;
	}
	return retval;
}

bool Dictionaries::suggestRemoveLetter(const char *active_word, char *dest1, char *dest2, char *dest3) {
	ASSERT(active_word != 0);
	ASSERT(dest1 != 0);
	ASSERT(dest2 != 0);
	ASSERT(dest3 != 0);
	
	*dest1 = '\0';
	*dest2 = '\0';
	*dest3 = '\0';
	
	char newword[100 + 1];
	bool retval = false;
	int i;
	int len = strlen(active_word);
	if (len > 100 || len == 0)
		return false;
	// warning: utf8 codes are messed up!!!
	// lucian: provide an implementation to suport variable char length for utf8
	for ( i = 0; i < len; i++) {
		memset(newword, 0, 100 + 1);
		strncpy(newword, active_word, i);
		ASSERT(len - i - 1 >= 0);
		strncpy(newword + i, active_word + i + 1, len - i - 1);

		if (isUtf8Valid(newword)) {
			if (hptr_->findString(newword, true) ||
			    hptr_->findString(newword)) {
				retval = true;
				if (*dest1 == '\0')
					strcpy(dest1, newword);
				else if (*dest2 == '\0' && strcmp(dest1, dest2) != 0)
					strcpy(dest2, newword);
				else if (*dest3 == '\0' && strcmp(dest1, dest3) != 0 && strcmp(dest2, dest3) != 0) {
					strcpy(dest3, newword);
					return true;
				}
			}
		}
	}
	return retval;
}

bool Dictionaries::suggestInsertTrylist(const char *active_word, char *dest1, char *dest2, char *dest3) {
	ASSERT(active_word != 0);
	ASSERT(dest1 != 0);
	ASSERT(dest2 != 0);
	ASSERT(dest3 != 0);

	if (trylist_ == 0) // trylist not specified in the dictionary
		return false;
	*dest1 = '\0';
	*dest2 = '\0';
	*dest3 = '\0';
	
	char newword[100 + 1];
	bool retval = false;
	int i, j;
	int len = strlen(active_word);
	int tlen = strlen(trylist_);
	if (len > 100 || len == 0)
		return false;

	// warning: utf8 codes are messed up!!!
	// lucian: provide an implementation to suport variable char length for utf8
	for ( i = 1; i < len; i++) {
		for (j = 0; j < tlen; j++) {
			if (trylist_[j] & 0x80) // skeep utf8 chars - lucian: to fix this
				continue;
			if (isupper(trylist_[j]))
				continue;
			memset(newword, 0, 100 + 1);
			strncpy(newword, active_word, i);
			newword[i] = trylist_[j];
			strncpy(newword + i + 1, active_word + i, len - i);

			if (isUtf8Valid(newword)) {
				if (hptr_->findString(newword, true) ||
				    hptr_->findString(newword)) {
					retval = true;
					if (*dest1 == '\0')
						strcpy(dest1, newword);
					else if (*dest2 == '\0' && strcmp(dest1, dest2) != 0)
						strcpy(dest2, newword);
					else if (*dest3 == '\0' && strcmp(dest1, dest3) != 0 && strcmp(dest2, dest3) != 0) {
						strcpy(dest3, newword);
						return true;
					}
				}
			}
		}
	}
	return retval;
}

bool Dictionaries::suggestReplaceTrylist(const char *active_word, char *dest1, char *dest2, char *dest3) {
	ASSERT(active_word != 0);
	ASSERT(dest1 != 0);
	ASSERT(dest2 != 0);
	ASSERT(dest3 != 0);

	if (trylist_ == 0) // trylist not specified in the dictionary
		return false;
	*dest1 = '\0';
	*dest2 = '\0';
	*dest3 = '\0';
	
	char newword[100 + 1];
	bool retval = false;
	int i, j;
	int len = strlen(active_word);
	int tlen = strlen(trylist_);
	if (len > 100 || len == 0)
		return false;

	// warning: utf8 codes are messed up!!!
	// lucian: provide an implementation to suport variable char length for utf8
	for ( i = 1; i < len; i++) {
		for (j = 0; j < tlen; j++) {
			if (trylist_[j] & 0x80) // skeep utf8 chars - lucian: to fix this
				continue;
			if (isupper(trylist_[j]))
				continue;
			memset(newword, 0, 100 + 1);
			strncpy(newword, active_word, i);
			newword[i] = trylist_[j];
			strncpy(newword + i + 1, active_word + i + 1, len - i - 1);

			if (isUtf8Valid(newword)) {
				if (hptr_->findString(newword, true) ||
				    hptr_->findString(newword)) {
					retval = true;
					if (*dest1 == '\0')
						strcpy(dest1, newword);
					else if (*dest2 == '\0' && strcmp(dest1, dest2) != 0)
						strcpy(dest2, newword);
					else if (*dest3 == '\0' && strcmp(dest1, dest3) != 0 && strcmp(dest2, dest3) != 0) {
						strcpy(dest3, newword);
						return true;
					}
				}
			}
		}
	}
	return retval;
}

bool Dictionaries::suggestAddEndTrylist(const char *active_word, char *dest1, char *dest2, char *dest3) {
	ASSERT(active_word != 0);
	ASSERT(dest1 != 0);
	ASSERT(dest2 != 0);
	ASSERT(dest3 != 0);

	if (trylist_ == 0) // trylist not specified in the dictionary
		return false;
	*dest1 = '\0';
	*dest2 = '\0';
	*dest3 = '\0';
	
	char newword[100 + 1];
	bool retval = false;
	int i, j;
	int len = strlen(active_word);
	int tlen = strlen(trylist_);
	if (len > 100 || len == 0)
		return false;

	// warning: utf8 codes are messed up!!!
	// lucian: provide an implementation to suport variable char length for utf8
	i = len;
	for (j = 0; j < tlen; j++) {
		if (trylist_[j] & 0x80) // skeep utf8 chars - lucian: to fix this
			continue;
		if (isupper(trylist_[j]))
			continue;
		memset(newword, 0, 100 + 1);
		strncpy(newword, active_word, i);
		newword[i] = trylist_[j];
		if (isUtf8Valid(newword)) {
			if (hptr_->findString(newword, true) ||
			    hptr_->findString(newword)) {
				retval = true;
				if (*dest1 == '\0')
					strcpy(dest1, newword);
				else if (*dest2 == '\0' && strcmp(dest1, dest2) != 0)
					strcpy(dest2, newword);
				else if (*dest3 == '\0' && strcmp(dest1, dest3) != 0 && strcmp(dest2, dest3) != 0) {
					strcpy(dest3, newword);
					return true;
				}
			}
		}
	}
	return retval;
}

static bool isUtf8Valid(const char *str) {
	unsigned char *ptrin = (unsigned char *) str;
	while (*ptrin != '\0') {
		// check utf8 chars
		int cnt = 1;
		int i;
		if ((*ptrin & 0x80) == 0)
			cnt = 1;
		else if ((*ptrin & 0xc0) == 0xc0)
			cnt = 2;
		else if ((*ptrin & 0xe0) == 0xe0)
			cnt = 3;
		else if ((*ptrin & 0xf0) == 0xf0)
			cnt = 4;
		else if ((*ptrin & 0xf8) == 0xf8)
			cnt = 5;
		else if ((*ptrin & 0xfc) == 0xfc)
			cnt = 6;
		else {
			return false;
		}

		for (i = 0; i < cnt; i++) {
			if (*(ptrin + i) == '\0') {
				return false;
			}
		}
		if (cnt >= 2) {
			for (i = 1; i < cnt; i++) {
				if ( (*(ptrin + i) & 0xc0) != 0x80) {
					return false;
				}
			}
		}
		ptrin += cnt;
	}
	return true;
}
