#ifndef PROFILER_H
#define PROFILER_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

//***************************************************
// configuration section
//***************************************************

#ifndef LOG
#define LOG printf
#endif

#ifndef LOG_ERROR
#define LOG_ERROR printf
#endif

class ProfConfig
{
	public:
		// maximum number of independent profiling points in the code
		static const int MAX_STORAGE = 100;

		// choose print/no_print profile data on program exit
		// default: print
		enum {
			NO_PRINT = 0,
			PRINT
		};
		static const int PRINT_ON_EXIT = PRINT;
};

//***************************************************
// implementation
//***************************************************

// profiler class
template <class CONFIG>
class ProfTemplate
{
	private:

		// local data for porfiler class
		struct ProfStorage
		{

			// sturcture holding a specific count
			struct Storage
			{
				// entry name, just a pointer, no new/delete here
				char *name_;
				// number of calls
				unsigned cnt_;
				// maximum time between start and stop
				unsigned max_;
				// minimum time between start and stop
				unsigned min_;
				// calculate an average
				long long total_;

				// clean
				void clean() {
					cnt_ = max_ = total_ =  0;
					min_ = -1;
				}

				// constructor
				Storage(): name_(0) {
					clean();
				}
			};

			// time conversion: number of timestamp counts in a millisecond
			float tscale_;
			// counts storage area
			Storage storage_[CONFIG::MAX_STORAGE];
			// current number of entries handled
			int max_index_;
			// costructor
			ProfStorage(): tscale_(0), max_index_(1) {
				// first index is not in use!!!
			}
			// implementing a simple Mayer singleton
			static ProfStorage *instance(bool del = false) {
				static ProfStorage *pstorage = 0;
				static bool released = false;

				if (del == true) {
					released = true;
				}

				if (del == false && released == false) {
					if (pstorage == 0) {
						pstorage = new ProfStorage();
						LOG("Init profiler atexit\n");
						atexit(ProfTemplate::release);
					}
				}
				else {
					if (pstorage != 0) {
						if (CONFIG::PRINT_ON_EXIT == CONFIG::PRINT)
							pstorage->print();
						delete pstorage;
						pstorage = 0;
					}
				}

				return pstorage;
			}
			float calibrate() {
				LOG(".");

				// do it again
				unsigned temp1 = clock();
				// wait next OS tick
				while (temp1 == (unsigned) clock())
					;
				temp1 = (unsigned) clock();
				long long tstamp1 = tstamp();

				unsigned temp2 = temp1;
				// wait next OS tick
				while (temp2 == (unsigned) clock())
					;
				temp2 = (unsigned) clock();
				long long tstamp2 = tstamp();
				unsigned delta1 = temp2 - temp1;
				unsigned delta2 = (unsigned) (tstamp2 - tstamp1);
				float fdelta1 = (float) delta1 / (float) CLOCKS_PER_SEC;
				fdelta1 *= 1000;  // convert to ms
				float delta = (float)delta2/fdelta1;

				return delta;
			}

			void print() {
				// calculate a time conversion scale
				if (tscale_ == 0) {
					LOG("calibrating...");
					tscale_ = calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ += calibrate();
					tscale_ /= 10;
					LOG("\nprofiler %f counts per millisecond\n", tscale_);
					LOG("CPU %dMHz\n", (int) tscale_ / 1000);
				}
				LOG("---------------------------------------------------------------------------\n");
				LOG("NAME                        CNT         MAX(ms)     MIN(ms)    AVERAGE(ms)\n");
				LOG("---------------------------------------------------------------------------\n");
				int i;
				for (i = 0; i < max_index_; i++) {
					typename ProfStorage::Storage *ptr = &storage_[i];
					// do not print a zero count entry
					if (ptr->cnt_ == 0)
						continue;

					char temp[80];
					int j;
					for (j = 0; j < 80; j++)
						temp[j] = '\0';
					temp[79] = '\0';
					for (j = 0; j < 29 && ptr->name_[j] != '\0'; j++)
						temp[j] = ptr->name_[j];

					float average = 0;
					if (ptr->cnt_ != 0) {
						average = (float) ptr->total_ / (float) ptr->cnt_;
						average /= tscale_;
					}
					sprintf(&temp[28], "%d", ptr->cnt_);
					sprintf(&temp[40], "%.4f", (float) ptr->max_ / tscale_);
					sprintf(&temp[52], "%.4f", (float) ptr->min_ / tscale_);
					sprintf(&temp[63], "%.4f", average);

					for (j = 0; j < 63; j++)
						if (temp[j] == '\0')
							temp[j] = ' ';
					LOG("%s\n", temp);
				}
				LOG("---------------------------------------------------------------------------\n");
			}
		};				  //ProfStorage

	public:
		// timing measurement function using some instruction counter
		// in CPU; this implementation is specific to gcc on x86;
		static long long tstamp() {
			// on gcc/x86 (Pentium and up) platforms
			long long retval;
			__asm__ __volatile__ ("rdtsc" : "=A" (retval));
			return retval;

			// on MSVC/x86 it could look something like this:
			// ulong32 dwLow, dwHigh;
			//__asm
			//{
			//	rdtsc
			//	mov dwLow, eax
			//	mov dwHigh, edx
			//}
			//return ( (uint64)dwHigh << 32 ) | (uint64)dwLow;

			// on HP-UX, Solaris and RTLinux use the high-res timer facility
			// (man gethrtime); it might also work on other Unixes
			//long long retval;
			//retval = gethrtime();
			//return retval;
		}

		// exit point called by OS; it was set above with atexit()
		static void release() {
			ProfStorage::instance(true);
		}

		// add a new profiler entry
		static int add(char *name) {
			ProfStorage *pstorage = ProfStorage::instance();
			if (pstorage == 0) {
				return 0;
			}

			if (pstorage->max_index_ >= CONFIG::MAX_STORAGE) {
				LOG_ERROR("profiler storage excedded\n");
				fflush(0);
				exit(1);
			}

			pstorage->storage_[pstorage->max_index_].name_ = name;
			pstorage->max_index_++;

			return pstorage->max_index_ - 1;
		}

		// clean the profile trace
		static void clean() {
			ProfStorage *pstorage = ProfStorage::instance();
			if (pstorage == 0) {
				return;
			}

			int i;
			for (i = 0; i < pstorage->max_index_; i++) {
				typename ProfStorage::Storage *ptr = &pstorage->storage_[i];
				ptr->clean();
			}
		}

		// print all profile counts
		static void print() {
			ProfStorage *pstorage = ProfStorage::instance();
			if (pstorage == 0) {
				return;
			}
			pstorage->print();
		}

	public:
		// constructor: save the timestamp
		ProfTemplate(int index): index_(index) {
			start_= tstamp();
		}

		// destructor: calculate how much time has passed;
		// update storage area
		~ProfTemplate() {
			long long end = tstamp();
			// if the difference is to big, it will fail here!
			long long ldelta = end - start_;
			if (ldelta > 0xffffffff)
				ldelta = 0xffffffff;
			unsigned delta = (unsigned) ldelta;

			ProfStorage *pstorage = ProfStorage::instance();
			if (pstorage == 0) {
				return;
			}

			// update profiler entry area
			if (delta > pstorage->storage_[index_].max_)
				pstorage->storage_[index_].max_ = delta;
			if (delta < pstorage->storage_[index_].min_)
				pstorage->storage_[index_].min_ = delta;
			pstorage->storage_[index_].total_ += delta;
			pstorage->storage_[index_].cnt_++;

		}
	private:
		// profiler entry index in the profiler entry array,
		// also used as an id for this entry
		int index_;
		// starting time
		long long start_;
};						  //ProfTemplate

typedef ProfTemplate<ProfConfig> Profiler;

#define PROFILE(X) static int pidx_ = 0; \
if (pidx_ == 0) \
pidx_ = Profiler::add((char *) X); \
Profiler paction_(pidx_);
#endif						  // PROFILER_H
