//****************************************************************************
// Copyright (C) 2005,  Lucian Constantin <struct_bylighting@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//****************************************************************************

#include "qwerty.h"
#include "briefedit.h"
#include "dialogs.h"
#include "statusfield.h"
#include "bufmgr.h"
#include "busycursor.h"
#include "qtutils.h"
#include <qapplication.h>
#include <qregexp.h>

void Editor::redo() {
	cleanMWords();
	e_->redo();

	// force update of cursor postition
	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	e_->setCursorPosition(para, index);
}


void Editor::undo() {
	cleanMWords();
	e_->undo();

	// force update of cursor postition
	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	e_->setCursorPosition(para, index);
}

void Editor::cut() {
	LOG("Editor cut");
	cleanMWords();
	e_->cut();
	// force update of cursor postition
	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	e_->setCursorPosition(para, index);
}
void Editor::copy() {
	cleanMWords();

	// try first to copy from the current tab window
	ASSERT(tw_current_ != 0);
	QString text = tw_current_->markedText();
	if (!(text.isNull() || text.isEmpty()))
		tw_current_->copy();

	// otherwise copy from edit window
	else
		e_->copy();

	// force update of cursor postition
	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	e_->setCursorPosition(para, index);
}

void Editor::paste() {
	cleanMWords();
	e_->paste();

	// force update of cursor postition
	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	e_->setCursorPosition(para, index);
}
void Editor::e_delete() {
	cleanMWords();
	e_->del();

	// force update of cursor postition
	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	e_->setCursorPosition(para, index);
}

void Editor::selectAll() {
	cleanMWords();
	e_->selectAll(true);
}

void Editor::toggleOverwrite() {
	bool status = e_->isOverwriteMode();
LOG("%d overwrite");
	e_->setOverwriteMode(!status);

	if (status)
		iow_->setText("Insert");
	else
		iow_->setText("Overwrite");
}

void Editor::startOfBuffer() {
	cleanMWords();
	e_->setCursorPosition(0, 0);
}

void Editor::endOfBuffer() {
	cleanMWords();
	int lines = e_->numLines();
	e_->setCursorPosition(lines - 1, 0);
}

void Editor::gotoLine() {
	cleanMWords();
	LineDialog *ld = new LineDialog();
	ld->setCaption(tr("Line"));
	if (!ld->exec())
		return;

	int line = ld->getLine();
	LOG("Editor::gotoline %d", line);
	delete ld;

	if (line > 0) {
		int lines = e_->numLines();
		if (line > lines)
			line = lines;
		e_->setCursorPosition(line - 1, 0);
	}
}

void Editor::gotoBookmark() {
	cleanMWords();
	BookmarkDialog *bd = new BookmarkDialog();
	bd->setCaption(tr("Bookmark"));
	if (!bd->exec())
		return;
	int bookmark = bd->getBookmark();
	LOG("Editor::gotoBookmark %d", bookmark);
	delete bd;

	gotoBookmarkCommon(bookmark);
}

void Editor::gotoBookmarkCommon(int bookmark) {
	if (bookmark > MAXBOOKMARKS || bookmark < 0)
		return;

	// change the buffer if necessary
	QString current_fname = getFname();
	if (current_fname != book_[bookmark].fname_) {
		bool found = false;
		int i;
		for (i = 0; i < MAXBUFMGR; i++) {
			QString bfname = b_->getFName(i);
			if (bfname.isNull())
				break;
			if (bfname.startsWith("*"))
				bfname = bfname.mid(1);
			if (book_[bookmark].fname_ == bfname) {
				found = true;
				break;
			}
		}
		if (!found) {
			msgScreen(tr("Bookmark not found and deleted"), 5000);
			book_[bookmark].x_ = 0;
			book_[bookmark].y_ = 0;
			book_[bookmark].fname_ = "";
			book_[bookmark].line_ = "";
			regenBookmarks();
			return;
		}

		BCursor busy;
		// save current buffer data
		storeCurentBuffer();

		// move to the buffer specified by the bookmark
		b_->setActiveIndex(i);

		// restore data
		if (using_aux_)
			changed_aux_ = true;
		else
			changed_aux_ = true;
		QString fname = b_->getActiveFName();
		setFname(fname, b_->getActiveFormat(), false);
		int line = b_->getActiveLine();
		int col = b_->getActiveCol();
		e_->setCursorPosition(line, col, false);
		e_->setText(b_->getActiveText());
		if (using_aux_)
			changed_aux_ = fname.startsWith("*");
		else
			changed_main_ = fname.startsWith("*");
		e_->setCursorPosition(line, col, false);

		redoBufMgrList();
	}


	int line = book_[bookmark].y_;
	int lines = e_->numLines();
	if (line > lines)
		line = lines - 1;
	e_->setCursorPosition(line, book_[bookmark].x_);
	e_->setFocus();

	// verify and readjust the line if necessary
	QString txt = e_->textLine(line);
	unsigned short bookmarker = QTUTILS::getBookmarker(bookmark);
	QString bookstr = QChar(bookmarker);

	if (!txt.startsWith(bookstr)) {
		int newline = 1;
		int newindex = 1;
		// search forward
		if (e_->find(bookstr, true, false, true, &newline, &newindex))
			;
		else {
			msgScreen(tr("Bookmark not found and deleted"), 5000);
			book_[bookmark].x_ = 0;
			book_[bookmark].y_ = 0;
			book_[bookmark].fname_ = "";
			book_[bookmark].line_ = "";
			regenBookmarks();
			return;
		}

		if (newindex != 0) {
			// looks like we got new characters in the first position in the stream
			e_->del();
			e_->insertAt(bookstr, newline, 0);
		}
		msgScreen(tr("Bookmark adjusted"), 2000);
		book_[bookmark].y_ = newline;
		e_->setCursorPosition(newline, book_[bookmark].x_);
		txt = e_->textLine(newline);
		book_[bookmark].line_ = txt.mid(1);
	}
	else {
		txt = e_->textLine(line);
		book_[bookmark].line_ = txt.mid(1);
	}
	regenBookmarks();
}

void Editor::setBookmark() {
	cleanMWords();
	BookmarkDialog *bd = new BookmarkDialog();
	bd->setCaption(tr("Bookmark"));
	if (!bd->exec())
		return;
	int bookmark = bd->getBookmark();
	LOG("Editor::setBookmark %d", bookmark);
	delete bd;

	setBookmarkCommon(bookmark);
}

void Editor::regenBookmarks() {
	// regenerate bookmark window
	QString book = "";
	for (int i = 0; i < MAXBOOKMARKS; i++) {
		if (!book_[i].fname_.isEmpty()) {
			char str[20];
			sprintf(str, "%d: ", i);
			book += QString(str) + book_[i].fname_ + ":\n\t" + book_[i].line_ + "\n";
		}
	}

	tw_bookmarks_->setText(book);
	tw_->setCurrentPage(ID_BOOKMARKS);
}

void Editor::setBookmarkCommon(int bookmark) {
	cleanMWords();
	if (bookmark > MAXBOOKMARKS)
		return;

	int para;
	int index;
	e_->getCursorPosition(&para, &index);

	// if we already have a bookmark here, clear it first
	QString txt = e_->textLine(para);
	for (int i = 0; i < MAXBOOKMARKS; i++) {
		QString bookstr = QChar(QTUTILS::getBookmarker(i));
		if (txt.startsWith(bookstr)) {
			msgScreen(tr("Old bookmark cleared"), 2000);
			book_[i].x_ = 0;
			book_[i].y_ = 0;
			book_[i].fname_ = "";
			book_[i].line_ = "";
			e_->setSelection(para, 0, para, 1);
			e_->del();
			break;
		}
	}


	QString current_fname = getFname();
	// clear the old bookmark if necessary
	if (!current_fname.isEmpty() && current_fname != "*") {

		if (current_fname == book_[bookmark].fname_) {
			storeCurentBuffer();
			QString txt = b_->getActiveText();
			txt = QTUTILS::cleanBookmarker(txt, bookmark);
			e_->setText(txt);
			storeCurentBuffer();
		}
		else {
			// process an inactive buffer
			int findex;
			if (b_->hasFile(book_[bookmark].fname_, &findex)) {
				QString txt = b_->getText(findex);
				txt = QTUTILS::cleanBookmarker(txt, bookmark);
				b_->setText(txt, findex);
			}
		}
	}
	else {
		msgScreen(tr("Cannot add bookmark in an unnamed file"), 5000);
		return;
	}

	book_[bookmark].x_ = index;
	book_[bookmark].y_ = para;
	book_[bookmark].fname_ = getFname();
	book_[bookmark].line_ = e_->textLine(para);

	// insert bookmarker
	QString bookstr = QChar(QTUTILS::getBookmarker(bookmark));
	e_->insertAt(bookstr, para, 0);
	regenBookmarks();
	e_->setCursorPosition(para, index);
}

void Editor::setBookmark1() {
	setBookmarkCommon(1);
}

void Editor::setBookmark2() {
	setBookmarkCommon(2);
}

void Editor::setBookmark3() {
	setBookmarkCommon(3);
}

void Editor::setBookmark4() {
	setBookmarkCommon(4);
}

void Editor::setBookmark5() {
	setBookmarkCommon(5);
}

void Editor::setBookmark6() {
	setBookmarkCommon(6);
}

void Editor::setBookmark7() {
	setBookmarkCommon(7);
}

void Editor::setBookmark8() {
	setBookmarkCommon(8);
}

void Editor::setBookmark9() {
	setBookmarkCommon(9);
}

void Editor::setBookmark0() {
	setBookmarkCommon(0);
}

void Editor::insertHex() {
	cleanMWords();
	HexDialog *hd = new HexDialog();
	hd->setCaption(tr("Unicode"));
	if (!hd->exec())
		return;
	int val = hd->getHex();
	if (val == -1)
		return;

	int para;
	int pos;
	e_->getCursorPosition(&para, &pos);
	QString newstr = QChar((unsigned short) val);
	e_->insertAt(newstr, para, pos);
	LOG("Editor::insertHex %x", val);
	delete hd;
}

void Editor::selectBlock() {
	int para;
	int index;
	int orig_para;
	int orig_index;
	
	e_->getCursorPosition(&para, &index);
	
	int lines = e_->numLines();
	if (para >= lines)
		return;

	QRegExp start_regex("[({\"]");

	// advance to the first position
	bool found = false;
	QString line = e_->textLine(para);
	while (para < lines) {
		line = e_->textLine(para);
//		index = line.find(tofind, index, !ignore_case);
		index = line.find(start_regex, index);
		if (index != -1) {
			found = true;
			break;
		}
		index = 0;
		para++;
	}
	if (!found) {
		msgScreen(tr("Cannot select block"), 5000);
		return;
	}
			
	orig_para = para;
	orig_index = index + 1;
	QString f1;
	QString f2;
	if (line.at(index) == '"') {
		f1 = "\"";
		f2 = "\"";
	}
	else if (line.at(index) == '{') {
		f1 = "{";
		f2 = "}";
	}
	else if (line.at(index) == '(') {
		f1 = "(";
		f2 = ")";
	}
	index++;
	int cnt = 1;

	// advance to the end position
	while (para < lines) {
		line = e_->textLine(para);
		if (line.contains(f1) != 0 || line.contains(f2) != 0) {
			// process the line character by character	
			while (index < (int) line.length()) {
				if (QString(line.at(index)) == f2)
					cnt--;
				else if (QString(line.at(index)) == f1)
					cnt++;
				if (cnt == 0)
					break;
				index++;
			}
		}
		if (cnt == 0)
			break;
		index = 0;
		para++;
	}
		
	e_->setCursorPosition(para, index, false);
	e_->setSelection(orig_para, orig_index, para, index);
	if (cnt != 0)
		msgScreen(tr("Cannot select block"), 5000);
}
