// Copyright (C) 2005,  Lucian Constantin <struct_bylighting@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#ifndef UTILS_H
#define UTILS_H

#include <string.h>
#include <ctype.h>
#include <sys/select.h> // POSIX 2001
#include "debug.h"

class UTILS {
	/* list definition for run_prg_malloc() */
	static const int PROG_BUF = 10000; // buffer size for run_prg_malloc
	static const int FCOMP_BUF = 512; // buffer size for file compare

	struct MyList {
		char buf[PROG_BUF + 1];
		MyList *next;

		MyList(): next(0) {
			memset(buf, 0, sizeof(buf));
		}

		~MyList() {
			if (next != 0) {
				delete next;
			}
		}
	};

public:
	static char *run_prg_malloc(const char *program) {
		PROFILE("UTILS::run_prg_malloc");
		FILE *input;
		MyList *ptr = new MyList();
		MyList *head = ptr;
		char *retval;
		int len;



		if (0 == (input = popen(program, "r"))) {
			fprintf(stderr, "rospell error: cannot run program %s\n", program);
			delete head;
			return 0;
		} else {
			while ((len = fread(ptr->buf, 1, PROG_BUF, input)) > 0) {
//				printf("Read %d bytes\n", len);
				ptr->next = new MyList();
				ptr = ptr->next;
			}

			pclose(input);
		}
		/* calculate length of return string */
		len = 0;
		ptr = head;
		while (ptr != 0) {
			len += strlen(ptr->buf);
			ptr = ptr->next;
		}

		if (len == 0) {
			delete head;
			return 0;
		}

		/* allocate retval +10 just in case... */
		if ((retval = new char[len + 10]) == 0) {
			delete head;
			return 0;
		}
		memset(retval, 0, len + 10);

		/* copy retval */
		ptr = head;
		while (ptr != 0) {
			strcat(retval, ptr->buf);
			ptr = ptr->next;
		}

		delete head;
		return retval;
	}


	static char *read_file_malloc(const char *name, bool err_print = true) {
		PROFILE("UTILS::read_file_malloc");
		FILE *input;
		int len;
		char *inbuf;
		ASSERT(name != 0);

		if ((input = fopen(name, "r")) == 0) {
			if (err_print)
				fprintf(stderr, "rospell error: cannot read file %s\n", name);
			return 0;
		}

		/* find the length of the file - fstat and stat are not exactely ANSI C */
		fseek(input, 0, SEEK_END);
		len = ftell(input);
		if (len == 0)
			return 0;

		/* alloc and clean */
		if ((inbuf = new char[len + 10/* just in case ... */])  == 0) {
			fclose(input);
			if (err_print)
				fprintf(stderr, "rospell error: cannot allocate memory for file %s\n", name);
			LOG("UTILS::read_file_malloc: Cannot alloc memory for %s\n", name);
			return 0;
		}
		memset(inbuf, 0, len + 10);

		/* rewind and read the file */
		rewind(input);

		fread(inbuf, 1, len, input);
		if (ferror(input)) {
			fclose(input);
			delete[] inbuf;
			if (err_print)
				fprintf(stderr, "rospell error: cannot read file %s\n", name);
			LOG("ferror %d", ferror(input));
			LOG("UTILS::read_file_malloc: Cannot read file %s", name);
			return 0;
		}

		fclose(input);
		return inbuf;
	}

	// save str string as file fname
	static bool write_file(const char *fname, const char *str) {
		PROFILE("UTILS::write_file");
		ASSERT(fname != 0);
		ASSERT(str != 0);

		FILE *fp = fopen(fname, "w");
		if (fp == 0) {
			fprintf(stderr, "rospell error: cannot open file %s\n", fname);
			LOG("Utils::write_file: cannot open  file %s", fname);
			return false;
		}

		fprintf(fp, "%s", str);
		fclose(fp);
		return true;
	}

	// replace src with dest in instr buffer; a new buffer is allocated
	static char *buffer_sed(char *instr, const char *src, const char *dest,
					    bool preserve_input = true) {

		PROFILE("UTILS::buffer_sed");

		int i = 0;
		char *ptr = instr;
		while ((ptr = strstr(ptr, src)) != 0 && *ptr != '\0') {
			ptr++;
			i++;
		}

		// save a copy of the input buffer
		char *incopy = 0;
		if (preserve_input) {
			incopy = new char[strlen(instr) + 1];
			strcpy(incopy, instr);
			// if no replacement to be done, just return
			if (i == 0)
				return incopy;
		}

		// allocate the output buffer
		int len = strlen(instr) + i * strlen(dest) + 10;
		char *retval = new char[len];
		memset(retval, 0, len);

		ptr = instr;
		char *oldptr = ptr;
		while ((ptr = strstr(ptr, src)) != 0 && *ptr != 0) {
			*ptr = '\0';
			strcat(retval, oldptr);
			strcat(retval, dest);
			oldptr = ptr + strlen(src);
			ptr = oldptr;
		}
		strcat(retval, oldptr);

		// put back instr
		if (preserve_input) {
			ASSERT(incopy != 0);
			strcpy(instr, incopy);
		}
		delete[] incopy;
		return retval;
	}

	static int file_compare(const char *file1, const char *file2) {
		ASSERT(file1 != 0 && file2 != 0);
		int retval = 0;
		FILE *in1;
		FILE *in2;
		char buf1[FCOMP_BUF];
		char buf2[FCOMP_BUF];
		int len1;
		int len2;

		if ((in1 = fopen(file1, "r")) == 0) {
			fprintf(stderr, "rospell error: cannot open file %s\n", file1);
			return -1;
		}
		if ((in2 = fopen(file2, "r")) == 0) {
			fprintf(stderr, "rospell error: cannot open file %s\n", file2);
			fclose(in1);
			return 1;
		}
		while (1) {
			len1 = fread(buf1, 1, FCOMP_BUF, in1);
			if (ferror(in1)) {
				fprintf(stderr, "rospell error: cannot read file %s\n", file1);
				retval = -1;
				break;
			}

			len2 = fread(buf2, 1, FCOMP_BUF, in2);
			if (ferror(in2)) {
				fprintf(stderr, "rospell error: cannot read file %s\n", file2);
				retval = 1;
				break;
			}
			ASSERT(len1 == len2);
			if (len1 == 0 && len2 == 0) {
				retval = 0;
				break;
			}
			if (len1 == 0) {
				retval = 1;
				break;
			}
			if (len2 == 0) {
				retval = -1;
				break;
			}

			if ((retval = memcmp(buf1, buf2, len1)) != 0) {
				break;
			}
		}
		fclose(in1);
		fclose(in2);
		return retval;
	}

	// select wrapper for a single fd and a timout specified
	static int selectTimeout(int fd, unsigned int timeout) {
		fd_set rfds;
		struct timeval tv;
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		tv.tv_sec = timeout;
		tv.tv_usec = 0;
		return select(fd + 1, &rfds, NULL, NULL, &tv);
	}


};

#endif //UTILS_H
